/**
 * @file
 * Defines JavaScript for the image_sizes module.
 */

(() => {
  customElements.define(
    "responsive-img",
    class extends HTMLElement {
      /** @type {ResizeObserver} - Observer for real image width. */
      #ro = new ResizeObserver(([entry]) => {
        ({ width: this.size } = entry.contentRect);
        this.update();
      });

      /** @type {number} - The current width of the element. */
      size = this.offsetWidth;

      /** @type {boolean} - Whether the image has been loaded. */
      loaded = false;

      /** @type {HTMLImageElement|null} - The image element. */
      img;

      /** @type {ShadowRoot|null} - Shadowroot */
      root;

      constructor() {
        super();
        const internals = this.attachInternals();
        this.root = internals.shadowRoot;

        if (!this.root) {
          this.root = this.attachShadow({ mode: "open" });
          const template = this.querySelector("template");
          this.root.appendChild(template.content.cloneNode(true));
        }

        this.img = this.shadowRoot.querySelector("img");
      }

      connectedCallback() {
        this.#ro.observe(this);
        this.update();
        this.img?.addEventListener("load", this.imgLoaded.bind(this));
      }

      disconnectedCallback() {
        this.#ro.disconnect();
        this.img?.removeEventListener("load", this.imgLoaded.bind(this));
      }

      /**
       * Update markup.
       * @returns {void}
       */
      update() {
        this.img.sizes = `${Math.round(this.size)}px`;
      }

      /**
       * onLoad Callback.
       * @returns {void}
       */
      imgLoaded() {
        if (this.loaded) return;
        this.loaded = this.img.complete;
        this.img.classList.add("loaded");
      }

      /**
       * Check if the element is in the viewport.
       * @returns {boolean}
       */
      inViewport() {
        const rect = this.getBoundingClientRect();
        return (
          rect.top >= 0 &&
          rect.left >= 0 &&
          rect.bottom <=
            (window.innerHeight || document.documentElement.clientHeight) &&
          rect.right <=
            (window.innerWidth || document.documentElement.clientWidth)
        );
      }
    }
  );
})();
