<?php

declare(strict_types=1);

namespace Drupal\image_sizes\FormDecorator;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\form_decorator\FormDecoratorBase;
use Drupal\responsive_image\ResponsiveImageStyleInterface;

/**
 * Plugin implementation for "responsive_image_style_form".
 *
 * @FormDecorator(
 *   hook = "form_responsive_image_style_form_alter"
 * )
 */
final class ResponsiveImageStyleFormAlter extends FormDecoratorBase {

  use StringTranslationTrait;
  use MessengerTrait;
  use LoggerChannelTrait;

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, ...$args) {
    $form = parent::buildForm($form, $form_state, ...$args);

    $triggering_element = $form_state->getTriggeringElement();
    if (
      isset($form['keyed_styles']['image_sizes.sizing']) ||
      ($triggering_element && $triggering_element['#value'] == 'image_sizes')
    ) {
      /** @var \Drupal\responsive_image\ResponsiveImageStyleInterface $responsive_image_style */
      $responsive_image_style = $this->inner->getEntity();
      $breakpoints = array_reverse(\Drupal::service('breakpoint.manager')->getBreakpointsByGroup($responsive_image_style->getBreakpointGroup()));

      foreach ($breakpoints as $breakpoint_id => $breakpoint) {
        foreach ($breakpoint->getMultipliers() as $multiplier) {
          // Use a container since we have a single breakpoint.
          $form['keyed_styles'][$breakpoint_id][$multiplier]['#type'] = 'container';

          // Reset image_mapping_type options since only the image_sizes option
          // makes sense.
          $form['keyed_styles'][$breakpoint_id][$multiplier]['image_mapping_type']['#type'] = 'hidden';
          $form['keyed_styles'][$breakpoint_id][$multiplier]['image_mapping_type']['#options'] = ['image_sizes' => $this->t('Select multiple image styles and use image sizes technology.')];

          // Make image_sizes the default.
          if ($form['keyed_styles'][$breakpoint_id][$multiplier]['image_mapping_type']['#default_value'] == '_none') {
            $form['keyed_styles'][$breakpoint_id][$multiplier]['image_mapping_type']['#default_value'] = 'image_sizes';
          }

          // Show image_style size mapping selection also for image_sizes.
          $form['keyed_styles'][$breakpoint_id][$multiplier]['sizes_image_styles']['#states']['visible'][] = 'or';
          $form['keyed_styles'][$breakpoint_id][$multiplier]['sizes_image_styles']['#states']['visible'][':input[name="keyed_styles[' . $breakpoint_id . '][' . $multiplier . '][image_mapping_type]"]'] = ['value' => 'image_sizes'];

          // Add preload selection to the form.
          $image_styles = image_style_options();
          $image_styles[ResponsiveImageStyleInterface::ORIGINAL_IMAGE] = $this->t('- None (original image) -');
          $image_styles[ResponsiveImageStyleInterface::EMPTY_IMAGE] = $this->t('- empty image -');
          $form['keyed_styles'][$breakpoint_id][$multiplier]['preload'] = [
            '#title' => $this->t('Preload style'),
            '#type' => 'select',
            '#default_value' => $responsive_image_style->getPreloadImageStyle(),
            '#options' => $image_styles,
            '#required' => TRUE,
            '#description' => $this->t('Select the image style you wish to use as a preload. It will be converted to base64 and displayed before the actual image is loaded.'),
          ];
        }
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\responsive_image\ResponsiveImageStyleInterface $responsive_image_style */
    $responsive_image_style = $this->inner->getEntity();
    // Remove all the existing mappings and replace with submitted values.
    $responsive_image_style->removeImageStyleMappings();
    if ($form_state->hasValue('keyed_styles')) {
      foreach ($form_state->getValue('keyed_styles') as $breakpoint_id => $multipliers) {
        foreach ($multipliers as $multiplier => $image_style_mapping) {
          // Create the mapping for image_sizes option.
          if ($image_style_mapping['image_mapping_type'] === 'image_sizes') {
            $mapping = [
              'image_mapping_type' => 'image_sizes',
              'image_mapping' => [
                'sizes_image_styles' => array_keys(array_filter($image_style_mapping['sizes_image_styles'])),
                'preload' => $image_style_mapping['preload'],
              ],
            ];
            $responsive_image_style->addImageStyleMapping($breakpoint_id, $multiplier, $mapping);
          }
        }
      }
    }

    return $this->inner->save($form, $form_state);
  }

}
