<?php

declare(strict_types=1);

namespace Drupal\image_sizes;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Image\ImageFactory;
use Drupal\Core\Template\Attribute;
use Drupal\file\FileInterface;
use Drupal\image\Entity\ImageStyle;
use Drupal\image\Plugin\Field\FieldType\ImageItem;
use Drupal\responsive_image\ResponsiveImageStyleInterface;

/**
 * Service for sizing images.
 *
 * @package Drupal\image_sizes
 */
class ImageSizesService implements ImageSizesServiceInterface {

  /**
   * Constructs the ImageSizesService service.
   *
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system service.
   * @param \Drupal\Core\Image\ImageFactory $imageFactory
   *   The image factory.
   *   The entity type manager service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The module handler.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator
   *   The file URL generator.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(
    protected readonly FileSystemInterface $fileSystem,
    protected readonly ImageFactory $imageFactory,
    protected readonly ModuleHandlerInterface $moduleHandler,
    protected readonly FileUrlGeneratorInterface $fileUrlGenerator,
    protected readonly EntityTypeManagerInterface $entityTypeManager
  ) {}

  public function buildAttributes(
    ResponsiveImageStyleInterface $responsive_image_style,
    ImageItem $item,
    bool $inline = FALSE,
    array|null $item_attributes = []
  ): array {
    /** @var FileInterface $file */
    $file = $item->entity;
    $uri = $file->getFileUri();

    $imageStyleIds = $responsive_image_style->getImageStyleIds();
    /** @var \Drupal\image\ImageStyleInterface $firstImageStyle */
    $firstImageStyle = $this->entityTypeManager->getStorage('image_style')->load(reset($imageStyleIds));

    // Determine the dimensions of the styled image.
    $dimensions = [
      'width' => $item->get('width')->getValue(),
      'height' => $item->get('height')->getValue(),
    ];

    $firstImageStyle->transformDimensions($dimensions, $uri);

    $item_attributes = new Attribute($item_attributes);
    $image_attributes = new Attribute([
      'src' => $this->buildUrl($uri, $responsive_image_style, 'fallback'),
      'srcset' => $this->buildSrcSet($this->buildStyles($responsive_image_style, $file)),
      'part' => 'image',
      'width' => $dimensions['width'],
      'height' => $dimensions['height'],
      ...$this->buildImageItemAttributes($item),
    ]);
    $placeholder_attributes = new Attribute();

    /** @var \Drupal\image\Entity\ImageStyle $preload_style */
    if ($inline === TRUE && ($preload_style = $this->entityTypeManager->getStorage('image_style')->load($responsive_image_style->getPreloadImageStyle()))) {
      $placeholder_attributes->merge($this->buildPlaceholderAttributes($preload_style, $file));
    }

    return [
      $item_attributes,
      $image_attributes,
      $placeholder_attributes,
    ];
  }

  /**
   * Builds attributes based on an image item.
   *
   * @param \Drupal\image\Plugin\Field\FieldType\ImageItem $item
   *   The image item.
   * @param $allowed_properties
   *   Allowed properties. All other properties are getting ignored from the
   *   item.
   *
   * @return array
   *   An attributes array.
   */
  public function buildImageItemAttributes(ImageItem $item, $allowed_properties = ['alt', 'title']): array {
    $attributes = $item->getValue();
    return array_filter($attributes,
      function ($key) use ($allowed_properties, $attributes) {
        return in_array($key, $allowed_properties) && !empty($attributes[$key]);
      }, ARRAY_FILTER_USE_KEY
    );
  }

  /**
   * Create base64 attributes based on an image and an image_style.
   *
   * @param \Drupal\image\Entity\ImageStyle $image_style
   *   The image style to use.
   * @param \Drupal\file\FileInterface $image
   *   The image to process.
   */
  public function buildPlaceholderAttributes(ImageStyle $image_style, FileInterface $image): Attribute {
    // Make sure to not give access through the base64 if not granted.
    if (!$image->access('view')) {
      return [];
    }

    $attributes = new Attribute();
    $attributes
      ->setAttribute('alt', 'placeholder')
      ->setAttribute('aria-hidden', true)
      ->setAttribute('part', 'placeholder');

    $image_uri = $image->getFileUri();
    $image_type = $image->getMimeType();
    $derivative_uri = $image_style->buildUri($image_uri);
    if (!file_exists($derivative_uri)) {
      $image_style->createDerivative($image_uri, $derivative_uri);
    }
    $absolute_path = $this->fileSystem->realpath($derivative_uri);
    if ($absolute_path && file_exists($absolute_path)) {
      $image_file = file_get_contents($absolute_path);
      $base_64_image = base64_encode($image_file);
      $attributes->setAttribute('src', "data:$image_type;base64,$base_64_image");

      $image_media = $this->imageFactory->get($absolute_path);
      if ($image_media->isValid()) {
        $attributes->setAttribute('width', $image_media->getWidth());
        $attributes->setAttribute('height', $image_media->getHeight());
      }
    }

    return $attributes;
  }

  /**
   * @param string $uri
   * @param ResponsiveImageStyleInterface $responsive_image_style
   * @param string $type
   * @return string
   */
  protected function buildUrl(string $uri, ResponsiveImageStyleInterface $responsive_image_style, string $type): string {
    if ($type == 'fallback') {
      if ($responsive_image_style->getFallbackImageStyle() == 'original') {
        return $this->fileUrlGenerator->generateAbsoluteString($uri);
      }
      /** @var \Drupal\image\Entity\ImageStyle $image_style */
      $image_style = $this->entityTypeManager->getStorage('image_style')->load($responsive_image_style->getFallbackImageStyle());
    }

    if ($type == 'preload') {
      /** @var \Drupal\image\Entity\ImageStyle $image_style */
      $image_style = $this->entityTypeManager->getStorage('image_style')->load($responsive_image_style->getPreloadImageStyle());
    }

    if (!$image_style) {
      return $this->fileUrlGenerator->generateAbsoluteString($uri);
    }

    return $image_style->buildUrl($uri);
  }

  /**
   * Build all the styles for an uri based on the responsive_image_style.
   *
   * @param string $uri
   *   The file uri to build the styles with.
   * @param \Drupal\responsive_image\ResponsiveImageStyleInterface $responsive_image_style
   *   The responsive image style.
   *
   * @return array
   *   An array of styles.
   */
  public function buildStyles(ResponsiveImageStyleInterface $responsive_image_style, FileInterface $file): array {
    $uri = $file->getFileUri();
    $image = $this->imageFactory->get($uri);
    $styles = [];

    /** @var \Drupal\image\ImageStyleInterface $image_style */
    foreach ($this->entityTypeManager->getStorage('image_style')->loadMultiple($responsive_image_style->getImageStyleIds()) as $image_style) {
      if ($image->isValid()) {
        $uri = $image->getSource();
        $dimensions = [
          'width' => $image->getWidth(),
          'height' => $image->getHeight(),
        ];
        $image_style->transformDimensions($dimensions, $uri);
        if ($dimensions['width'] > 0 && !isset($styles[(string)$dimensions['width']])) {
          $dimensions['width'] = round($dimensions['width']);
          $styles[(string)$dimensions['width']] = $image_style->buildUrl($uri);
        }
      }
      // Try getting the information from the image style as a fallback.
      else {
        $effects = $image_style->getEffects();
        foreach ($effects as $effect) {
          if (is_a($effect, 'Drupal\image\Plugin\ImageEffect\ResizeImageEffect')) {
            $width = $effect->getConfiguration()['data']['width'];
            if ($width !== NULL && $width > 0) {
              $styles[(string)$width] = $image_style->buildUrl($uri);
            }
          }
        }
      }
    }

    ksort($styles);
    return $styles;
  }

  /**
   * Build a srcset string based on the styles.
   *
   * @param array $styles
   *   The styles to build the srcset with.
   *
   * @return string
   *   The srcset string.
   */
  public function buildSrcSet(array $styles): string {
    return array_reduce(array_keys($styles), function ($a, $c) use ($styles) {
      return "{$a} $styles[$c] {$c}w,";
    }, '');
  }

}
